@@ -78,6 +78,8 @@ WCHAR **main_wargv = NULL;
 
 static LCID user_lcid, system_lcid;
 static LANGID user_ui_language;
+static WCHAR *nt_build_dir;
+static WCHAR *nt_data_dir;
 
 static char system_locale[LOCALE_NAME_MAX_LENGTH];
 static char user_locale[LOCALE_NAME_MAX_LENGTH];
@@ -147,11 +149,12 @@ static void *read_nls_file( const char *name )
 
 static NTSTATUS open_nls_data_file( const char *path, const WCHAR *sysdir, HANDLE *file )
 {
-    NTSTATUS status = STATUS_OBJECT_NAME_NOT_FOUND;
+    NTSTATUS status;
+    IO_STATUS_BLOCK io;
     OBJECT_ATTRIBUTES attr;
     UNICODE_STRING valueW;
     WCHAR buffer[64];
-    char *p, *ntpath;
+    char *p;
 
     wcscpy( buffer, system_dir );
     p = strrchr( path, '/' ) + 1;
@@ -163,11 +166,7 @@ static NTSTATUS open_nls_data_file( const char *path, const WCHAR *sysdir, HANDL
                              FILE_OPEN, FILE_SYNCHRONOUS_IO_ALERT, NULL, 0 );
     if (status != STATUS_NO_SUCH_FILE) return status;
 
-    if ((status = nt_to_unix_file_name( &attr, &ntpath, FILE_OPEN ))) return status;
-    status = open_unix_file( file, ntpath, GENERIC_READ, &attr, 0, FILE_SHARE_READ,
-                             FILE_OPEN, FILE_SYNCHRONOUS_IO_ALERT, NULL, 0 );
-    free( ntpath );
-    return status;
+    return NtOpenFile( file, GENERIC_READ, &attr, &io, FILE_SHARE_READ, FILE_SYNCHRONOUS_IO_ALERT );
 }
 
 static NTSTATUS get_nls_section_name( UINT type, UINT id, WCHAR name[32] )
@@ -321,21 +320,13 @@ static void init_unix_codepage(void)
 #endif  /* __APPLE__ || __ANDROID__ */
 
 
-static inline SIZE_T get_env_length( const WCHAR *env )
-{
-    const WCHAR *end = env;
-    while (*end) end += wcslen(end) + 1;
-    return end + 1 - env;
-}
-
-
 #define STARTS_WITH(var,str) (!strncmp( var, str, sizeof(str) - 1 ))
 
 /***********************************************************************
  *           is_special_env_var
  *
  * Check if an environment variable needs to be handled specially when
- * passed through the Unix environment (i.e. prefixed with "WINE").
+ * imported from the Unix environment (i.e. prefixed with "WINE_HOST_").
  */
 static BOOL is_special_env_var( const char *var )
 {
@@ -344,29 +335,30 @@ static BOOL is_special_env_var( const char *var )
             STARTS_WITH( var, "HOME=" ) ||
             STARTS_WITH( var, "TEMP=" ) ||
             STARTS_WITH( var, "TMP=" ) ||
-            STARTS_WITH( var, "QT_" ) ||
-            STARTS_WITH( var, "VK_" ) ||
-            STARTS_WITH( var, "XDG_SESSION_TYPE=" ));
+            STARTS_WITH( var, "TMPDIR=" ) ||
+            STARTS_WITH( var, "XDG_" ));
 }
 
-/* check if an environment variable changes dynamically in every new process */
-static BOOL is_dynamic_env_var( const char *var )
+
+/***********************************************************************
+ *           is_ignored_env_var
+ *
+ * Check if an environment variable needs to be skipped when importing
+ * from the Unix environment, either because it would confuse Windows
+ * apps, or make the Windows environment grow too large.
+ */
+static BOOL is_ignored_env_var( const char *var )
 {
-    return (STARTS_WITH( var, "WINEDLLOVERRIDES=" ) ||
-            STARTS_WITH( var, "WINEDATADIR=" ) ||
-            STARTS_WITH( var, "WINEHOMEDIR=" ) ||
-            STARTS_WITH( var, "WINEBUILDDIR=" ) ||
-            STARTS_WITH( var, "WINECONFIGDIR=" ) ||
-            STARTS_WITH( var, "WINELOADER=" ) ||
-            STARTS_WITH( var, "WINEDLLDIR" ) ||
-            STARTS_WITH( var, "WINEUNIXCP=" ) ||
-            STARTS_WITH( var, "WINEUSERLOCALE=" ) ||
-            STARTS_WITH( var, "WINEUSERNAME=" ) ||
-            STARTS_WITH( var, "WINEPRELOADRESERVE=" ) ||
-            STARTS_WITH( var, "WINELOADERNOEXEC=" ) ||
-            STARTS_WITH( var, "WINESERVERSOCKET=" ));
+    return (STARTS_WITH( var, "NIXPKGS_" ) ||
+            STARTS_WITH( var, "QT_" ) ||
+            STARTS_WITH( var, "SDL_AUDIODRIVER=" ) ||
+            STARTS_WITH( var, "SDL_AUDIO_DRIVER=" ) ||
+            STARTS_WITH( var, "SDL_VIDEODRIVER=" ) ||
+            STARTS_WITH( var, "SDL_VIDEO_DRIVER=" ) ||
+            STARTS_WITH( var, "VK_" ));
 }
 
+
 /******************************************************************
  *      ntdll_umbstowcs  (ntdll.so)
  *
@@ -470,86 +462,18 @@ int ntdll_wcsnicmp( const WCHAR *str1, const WCHAR *str2, int n )
 /***********************************************************************
  *           ntdll_get_build_dir  (ntdll.so)
  */
-const char *ntdll_get_build_dir(void)
+const WCHAR *ntdll_get_build_dir(void)
 {
-    return build_dir;
+    return nt_build_dir;
 }
 
 
 /***********************************************************************
  *           ntdll_get_data_dir  (ntdll.so)
  */
-const char *ntdll_get_data_dir(void)
+const WCHAR *ntdll_get_data_dir(void)
 {
-    return data_dir;
-}
-
-
-/***********************************************************************
- *           build_envp
- *
- * Build the environment of a new child process.
- */
-char **build_envp( const WCHAR *envW )
-{
-    static const char * const unix_vars[] = { "PATH", "TEMP", "TMP", "HOME" };
-    char **envp;
-    char *env, *p;
-    int count = 1, length, lenW;
-    unsigned int i;
-
-    lenW = get_env_length( envW );
-    if (!(env = malloc( lenW * 3 ))) return NULL;
-    length = ntdll_wcstoumbs( envW, lenW, env, lenW * 3, FALSE );
-
-    for (p = env; *p; p += strlen(p) + 1, count++)
-    {
-        if (is_dynamic_env_var( p )) continue;
-        if (is_special_env_var( p )) length += 4; /* prefix it with "WINE" */
-    }
-
-    for (i = 0; i < ARRAY_SIZE( unix_vars ); i++)
-    {
-        if (!(p = getenv(unix_vars[i]))) continue;
-        length += strlen(unix_vars[i]) + strlen(p) + 2;
-        count++;
-    }
-
-    if ((envp = malloc( count * sizeof(*envp) + length )))
-    {
-        char **envptr = envp;
-        char *dst = (char *)(envp + count);
-
-        /* some variables must not be modified, so we get them directly from the unix env */
-        for (i = 0; i < ARRAY_SIZE( unix_vars ); i++)
-        {
-            if (!(p = getenv( unix_vars[i] ))) continue;
-            *envptr++ = strcpy( dst, unix_vars[i] );
-            strcat( dst, "=" );
-            strcat( dst, p );
-            dst += strlen(dst) + 1;
-        }
-
-        /* now put the Windows environment strings */
-        for (p = env; *p; p += strlen(p) + 1)
-        {
-            if (*p == '=') continue;  /* skip drive curdirs, this crashes some unix apps */
-            if (is_dynamic_env_var( p )) continue;
-            if (is_special_env_var( p ))  /* prefix it with "WINE" */
-            {
-                *envptr++ = strcpy( dst, "WINE" );
-                strcat( dst, p );
-            }
-            else
-            {
-                *envptr++ = strcpy( dst, p );
-            }
-            dst += strlen(dst) + 1;
-        }
-        *envptr = 0;
-    }
-    free( env );
-    return envp;
+    return nt_data_dir;
 }
 
 
@@ -912,6 +836,20 @@ void init_environment(void)
 }
 
 
+/* check if a WINE_HOST_ prefixed variable already exists in the environment */
+static BOOL host_var_exists( const char *name )
+{
+    char *end = strchr( name, '=' );
+
+    if (!end) return FALSE;
+    for (char **e = environ; *e; e++)
+    {
+        if (!STARTS_WITH( *e, "WINE_HOST_" )) continue;
+        if (!strncmp( *e + 10, name, end + 1 - name )) return TRUE;
+    }
+    return FALSE;
+}
+
 static const char overrides_help_message[] =
     "Syntax:\n"
     "  WINEDLLOVERRIDES=\"entry;entry;entry...\"\n"
@@ -935,7 +873,7 @@ static WCHAR *get_initial_environment( SIZE_T *pos, SIZE_T *size )
 
     /* estimate needed size */
     *size = 1;
-    for (e = environ; *e; e++) *size += strlen(*e) + 1;
+    for (e = environ; *e; e++) *size += strlen(*e) + 6;
 
     env = malloc( *size * sizeof(WCHAR) );
     ptr = env;
@@ -945,18 +883,24 @@ static WCHAR *get_initial_environment( SIZE_T *pos, SIZE_T *size )
         char *str = *e;
 
         /* skip Unix special variables and use the Wine variants instead */
-        if (!strncmp( str, "WINE", 4 ))
+        if (STARTS_WITH( str, "WINE" ))
         {
-            if (is_special_env_var( str + 4 )) str += 4;
+            if (is_special_env_var( str + 4 ) || is_ignored_env_var( str + 4 )) str += 4;
             else if (!strcmp( str, "WINEDLLOVERRIDES=help" ))
             {
                 MESSAGE( overrides_help_message );
                 exit(0);
             }
         }
-        else if (is_special_env_var( str )) continue;  /* skip it */
+        else if (is_ignored_env_var( str )) continue;
+        else if (host_var_exists( str )) continue;
+        else if (is_special_env_var( str )) /* prefix it with WINE_HOST_ */
+        {
+            static const WCHAR hostW[] = {'W','I','N','E','_','H','O','S','T','_'};
+            memcpy( ptr, hostW, sizeof(hostW) );
+            ptr += ARRAY_SIZE(hostW);
+        }
 
-        if (is_dynamic_env_var( str )) continue;
         ptr += ntdll_umbstowcs( str, strlen(str) + 1, ptr, end - ptr );
     }
     *pos = ptr - env;
@@ -1042,7 +986,7 @@ static void add_path_var( WCHAR **env, SIZE_T *pos, SIZE_T *size, const char *na
 {
     WCHAR *nt_name = NULL;
 
-    if (path && unix_to_nt_file_name( path, &nt_name )) return;
+    if (path && unix_to_nt_file_name( path, &nt_name, FILE_OPEN )) return;
     append_envW( env, pos, size, name, nt_name );
     free( nt_name );
 }
@@ -1058,7 +1002,7 @@ static void add_system_dll_path_var( WCHAR **env, SIZE_T *pos, SIZE_T *size )
     {
         WCHAR *nt_name = NULL;
 
-        if (!unix_to_nt_file_name( system_dll_paths[i], &nt_name ))
+        if (!unix_to_nt_file_name( system_dll_paths[i], &nt_name, FILE_OPEN ))
         {
             size_t len = wcslen( nt_name );
             path = realloc( path, (path_len + len + 1) * sizeof(WCHAR) );
@@ -1084,13 +1028,15 @@ static void add_system_dll_path_var( WCHAR **env, SIZE_T *pos, SIZE_T *size )
  */
 static void add_dynamic_environment( WCHAR **env, SIZE_T *pos, SIZE_T *size )
 {
-    const char *overrides = getenv( "WINEDLLOVERRIDES" );
     unsigned int i;
     char str[22];
 
-    add_path_var( env, pos, size, "WINEDATADIR", data_dir );
+    if (build_dir) unix_to_nt_file_name( build_dir, &nt_build_dir, FILE_OPEN );
+    if (data_dir) unix_to_nt_file_name( data_dir, &nt_data_dir, FILE_OPEN );
+
+    append_envW( env, pos, size, "WINEBUILDDIR", nt_build_dir );
+    append_envW( env, pos, size, "WINEDATADIR", nt_data_dir );
     add_path_var( env, pos, size, "WINEHOMEDIR", home_dir );
-    add_path_var( env, pos, size, "WINEBUILDDIR", build_dir );
     add_path_var( env, pos, size, "WINECONFIGDIR", config_dir );
     add_path_var( env, pos, size, "WINELOADER", wineloader );
     for (i = 0; dll_paths[i]; i++)
@@ -1102,7 +1048,6 @@ static void add_dynamic_environment( WCHAR **env, SIZE_T *pos, SIZE_T *size )
     append_envW( env, pos, size, str, NULL );
     add_system_dll_path_var( env, pos, size );
     append_envA( env, pos, size, "WINEUSERNAME", user_name );
-    append_envA( env, pos, size, "WINEDLLOVERRIDES", overrides );
     if (unix_cp.CodePage != CP_UTF8)
     {
         snprintf( str, sizeof(str), "%u", unix_cp.CodePage );
@@ -1476,7 +1421,7 @@ static WCHAR *get_initial_directory(void)
 
     if (pwd)
     {
-        if (!unix_to_nt_file_name( pwd, &ret ))
+        if (!unix_to_nt_file_name( pwd, &ret, FILE_OPEN ))
         {
             ULONG len = wcslen( ret );
             if (len && ret[len - 1] != '\\')
@@ -1620,6 +1565,7 @@ static void run_wineboot( WCHAR *env, SIZE_T size )
         return;
     }
 
+    virtual_init_user_shared_data();
     env[size] = 0;
     params.Flags           = PROCESS_PARAMS_FLAG_NORMALIZED;
     params.Environment     = env;
@@ -1655,7 +1601,7 @@ static void run_wineboot( WCHAR *env, SIZE_T size )
 
 wait:
     timeout.QuadPart = (ULONGLONG)5 * 60 * 1000 * -10000;
-    if (NtWaitForMultipleObjects( count, handles, TRUE, FALSE, &timeout ) == WAIT_TIMEOUT)
+    if (NtWaitForMultipleObjects( count, handles, WaitAny, FALSE, &timeout ) == WAIT_TIMEOUT)
         ERR( "boot event wait timed out\n" );
     while (count) NtClose( handles[--count] );
 }
@@ -1677,6 +1623,30 @@ static inline void put_unicode_string( WCHAR *src, WCHAR **dst, UNICODE_STRING *
     copy_unicode_string( &src, dst, str, wcslen(src) * sizeof(WCHAR) );
 }
 
+static void copy_dos_path_string( WCHAR **src, WCHAR **dst, UNICODE_STRING *str,
+                                  UNICODE_STRING *nt_str, UINT len )
+{
+    /* copy the original string into nt_str */
+    nt_str->Buffer = malloc( len + sizeof(WCHAR) );
+    memcpy( nt_str->Buffer, *src, len );
+    nt_str->Buffer[len / sizeof(WCHAR)] = 0;
+    nt_str->Length = len;
+    nt_str->MaximumLength = len + sizeof(WCHAR);
+
+    if (len > 5 * sizeof(WCHAR) && (*src)[5] == ':') /* skip the \??\ prefix */
+    {
+        *src += 4;
+        len -= 4 * sizeof(WCHAR);
+        copy_unicode_string( src, dst, str, len );
+    }
+    else
+    {
+        WCHAR *ptr = *dst;
+        copy_unicode_string( src, dst, str, len );
+        ptr[1] = '\\'; /* change \??\ to \\?\ */
+    }
+}
+
 static inline WCHAR *get_dos_path( WCHAR *nt_path )
 {
     if (nt_path[4] && nt_path[5] == ':') return nt_path + 4; /* skip the \??\ prefix */
@@ -1923,9 +1893,10 @@ static RTL_USER_PROCESS_PARAMETERS *build_initial_params( void **module )
     static const WCHAR pathW[] = {'P','A','T','H'};
     RTL_USER_PROCESS_PARAMETERS *params = NULL;
     SIZE_T size, env_pos, env_size;
-    WCHAR *dst, *image, *cmdline, *path, *bootstrap;
+    WCHAR *dst, *cmdline, *path, *bootstrap;
     WCHAR *env = get_initial_environment( &env_pos, &env_size );
     WCHAR *curdir = get_initial_directory();
+    UNICODE_STRING nt_name;
     NTSTATUS status;
 
     if (NtCurrentTeb64()) NtCurrentTeb64()->TlsSlots[WOW64_TLS_FILESYSREDIR] = TRUE;
@@ -1949,7 +1920,16 @@ static RTL_USER_PROCESS_PARAMETERS *build_initial_params( void **module )
     add_registry_environment( &env, &env_pos, &env_size );
     env[env_pos++] = 0;
 
-    status = load_main_exe( NULL, main_argv[1], curdir, 0, &image, module );
+    get_full_path( main_argv[1], curdir, &nt_name );
+    status = load_main_exe( &nt_name, 0, module );
+    /* fail only if the file contained an explicit path */
+    if (status == STATUS_DLL_NOT_FOUND &&
+        (strpbrk( main_argv[1], "/\\" ) || (main_argv[1][0] && main_argv[1][1] == ':')))
+    {
+        MESSAGE( "wine: failed to open %s\n", debugstr_a(main_argv[1]) );
+        NtTerminateProcess( GetCurrentProcess(), status );
+    }
+
     if (NT_SUCCESS(status))
     {
         char *loader;
@@ -1966,9 +1946,9 @@ static RTL_USER_PROCESS_PARAMETERS *build_initial_params( void **module )
     if (status)  /* try launching it through start.exe */
     {
         static const char *args[] = { "start.exe", "/exec" };
-        free( image );
+        free( nt_name.Buffer );
         if (*module) NtUnmapViewOfSection( GetCurrentProcess(), *module );
-        load_start_exe( &image, module );
+        load_start_exe( &nt_name, module );
         prepend_argv( args, 2 );
     }
     else
@@ -1977,7 +1957,7 @@ static RTL_USER_PROCESS_PARAMETERS *build_initial_params( void **module )
         if (NtCurrentTeb64()) NtCurrentTeb64()->TlsSlots[WOW64_TLS_FILESYSREDIR] = FALSE;
     }
 
-    main_wargv = build_wargv( get_dos_path( image ));
+    main_wargv = build_wargv( get_dos_path( nt_name.Buffer ));
     cmdline = build_command_line( main_wargv );
 
     TRACE( "image %s cmdline %s dir %s\n",
@@ -2008,7 +1988,7 @@ static RTL_USER_PROCESS_PARAMETERS *build_initial_params( void **module )
     put_unicode_string( main_wargv[0], &dst, &params->ImagePathName );
     put_unicode_string( cmdline, &dst, &params->CommandLine );
     put_unicode_string( main_wargv[0], &dst, &params->WindowTitle );
-    free( image );
+    free( nt_name.Buffer );
     free( cmdline );
     free( curdir );
 
@@ -2028,12 +2008,13 @@ static RTL_USER_PROCESS_PARAMETERS *build_initial_params( void **module )
  */
 void init_startup_info(void)
 {
-    WCHAR *src, *dst, *env, *image;
+    WCHAR *src, *dst, *env;
     void *module = NULL;
     unsigned int status;
     SIZE_T size, info_size, env_size, env_pos;
     RTL_USER_PROCESS_PARAMETERS *params = NULL;
     struct startup_info_data *info;
+    UNICODE_STRING nt_name;
     USHORT machine;
 
     if (!startup_info_size)
@@ -2107,7 +2088,7 @@ void init_startup_info(void)
     dst = params->CurrentDirectory.DosPath.Buffer + MAX_PATH;
 
     if (info->dllpath_len) copy_unicode_string( &src, &dst, &params->DllPath, info->dllpath_len );
-    copy_unicode_string( &src, &dst, &params->ImagePathName, info->imagepath_len );
+    copy_dos_path_string( &src, &dst, &params->ImagePathName, &nt_name, info->imagepath_len );
     copy_unicode_string( &src, &dst, &params->CommandLine, info->cmdline_len );
     copy_unicode_string( &src, &dst, &params->WindowTitle, info->title_len );
     copy_unicode_string( &src, &dst, &params->Desktop, info->desktop_len );
@@ -2129,16 +2110,15 @@ void init_startup_info(void)
     free( env );
     free( info );
 
-    status = load_main_exe( params->ImagePathName.Buffer, NULL, params->CommandLine.Buffer,
-                            machine, &image, &module );
+    status = load_main_exe( &nt_name, machine, &module );
     if (!NT_SUCCESS(status))
     {
-        MESSAGE( "wine: failed to start %s\n", debugstr_us(&params->ImagePathName) );
+        MESSAGE( "wine: failed to start %s: %x\n", debugstr_us(&params->ImagePathName), status );
         NtTerminateProcess( GetCurrentProcess(), status );
     }
     rebuild_argv();
-    main_wargv = build_wargv( get_dos_path( image ));
-    free( image );
+    main_wargv = build_wargv( params->ImagePathName.Buffer );
+    free( nt_name.Buffer );
     init_peb( params, module );
 }
 
@@ -2161,17 +2141,13 @@ void *create_startup_info( const UNICODE_STRING *nt_image, ULONG process_flags,
                            const struct pe_image_info *pe_info, DWORD *info_size )
 {
     struct startup_info_data *info;
-    UNICODE_STRING dos_image = *nt_image;
     DWORD size;
     void *ptr;
 
-    dos_image.Buffer = get_dos_path( nt_image->Buffer );
-    dos_image.Length = nt_image->Length - (dos_image.Buffer - nt_image->Buffer) * sizeof(WCHAR);
-
     size = sizeof(*info);
     size += params->CurrentDirectory.DosPath.Length;
     size += params->DllPath.Length;
-    size += dos_image.Length;
+    size += nt_image->Length;
     size += params->CommandLine.Length;
     size += params->WindowTitle.Length;
     size += params->Desktop.Length;
@@ -2210,7 +2186,7 @@ void *create_startup_info( const UNICODE_STRING *nt_image, ULONG process_flags,
     ptr = info + 1;
     info->curdir_len = append_string( &ptr, params, &params->CurrentDirectory.DosPath );
     info->dllpath_len = append_string( &ptr, params, &params->DllPath );
-    info->imagepath_len = append_string( &ptr, params, &dos_image );
+    info->imagepath_len = append_string( &ptr, params, nt_image );
     info->cmdline_len = append_string( &ptr, params, &params->CommandLine );
     info->title_len = append_string( &ptr, params, &params->WindowTitle );
     info->desktop_len = append_string( &ptr, params, &params->Desktop );
